# cython: language_level=3

from .messages import Msg

from . cimport zyre as z


# Zyre message grammar:
#     ENTER peer name headers ipaddress:port
#         a new peer has entered the network
#     EVASIVE peer name
#         a peer is being evasive (quiet for too long)
#     EXIT peer name
#         a peer has left the network
#     JOIN peer name group
#         a peer has joined a specific group
#     LEAVE peer name group
#         a peer has joined a specific group
#     WHISPER peer name message
#         a peer has sent this node a message
#     SHOUT peer name group message
#         a peer has sent one of our groups a message


MSG_SLOTS = {
    'ENTER': ('peer', 'name', 'headers', 'address'),
    'EVASIVE': ('peer', 'name'),
    'EXIT': ('peer', 'name'),
    'JOIN': ('peer', 'name', 'group'),
    'LEAVE': ('peer', 'name', 'group'),
    'WHISPER': ('peer', 'name', 'blob'),
    'SHOUT': ('peer', 'name', 'group', 'blob')
}
BIN_SLOTS = ('blob',)


cdef set zlist_to_str_set(z.zlist_t* zlist):
    """
    Convert a zlist to a set of str objects.
    
    Destroys the original zlist. 
    """
    return {s.decode('utf8') for s in zlist_to_bytes_set(zlist)}


cdef set zlist_to_bytes_set(z.zlist_t* zlist):
    """
    Convert a zlist to a set of bytes objects.
    
    Destroys the original zlist. 
    """
    cdef object py_set = set()
    cdef void* item = NULL;
    while z.zlist_size(zlist):
        item = z.zlist_pop(zlist)
        b_item = b'%s' % <char*>item
        py_set.add(b_item)
    z.zlist_destroy(&zlist)
    return py_set


cdef object zmsg_to_msg(z.zmsg_t *zmsg):
    """
    Convert a zmsg to a Msg instance.
    
    Destroys the original zmg.
    """
    cdef char * item
    try:
        if not z.zmsg_size(zmsg):
            raise ValueError('Invalid message')
        item = z.zmsg_popstr(zmsg)
        event = item.decode('utf8')
        parts = {'event': event}
        event = item.decode('utf8')
        parts = {'event': event}
        for slot in MSG_SLOTS[event]:
            if not z.zmsg_size(zmsg):
                raise ValueError('Invalid message')
            item = z.zmsg_popstr(zmsg)
            if slot not in BIN_SLOTS:
                parts[slot] = item.decode('utf8')
            else:
                parts[slot] = item
        msg =  Msg(**parts)
        return msg
    finally:
        z.zmsg_destroy(&zmsg)


# cython: language_level=3

import asyncio
import logging
import queue
import sys
import threading


from . import messages
from .exceptions import StartFailed, StopFailed, Stopped


# Importing cython.parallel ensures CPython's thread state is initialized properly
# See https://bugs.python.org/issue20891 and https://github.com/python/cpython/pull/5425
# and https://github.com/opensocdebug/osd-sw/issues/37
cimport cython.parallel

from cpython.ref cimport Py_INCREF, Py_DECREF
from libc.stdlib cimport free
from libc.string cimport strcmp

from . import futures
from . import nodeconfig
from . cimport signals
from . cimport util
from . cimport zyre as z


logger = logging.getLogger('aiozyre')


cdef class NodeActor:
    def __cinit__(
        self,
        *,
        config: nodeconfig.NodeConfig,
        loop: asyncio.AbstractEventLoop
    ):
        self.config = config
        self.loop = loop
        self.zactor_pipe = NULL
        self.zactor = NULL
        self.zpoller = NULL
        self.zyre = NULL
        self.zthreadid = -1
        self.lthreadid = threading.get_ident()
        self.started = None
        self.stopped = None

        # Use a non-thread-safe awaitable queue for sending messages from the zactor thread.
        # We achieve thread safety by using loop.call_soon_threadsafe to place
        # items in the queue from the zactor thread.
        self.outbox = asyncio.Queue()

        # Use a thread-safe queue, non-awaitable queue for sending messages to the zactor thread.
        # The zactor thread can't async/await since it is a cdef function, and since it is a separate thread
        # it is fine for it to block while waiting for a message.
        self.inbox = queue.Queue()

    def __init__(
        self,
        *,
        config: nodeconfig.NodeConfig,
        loop: asyncio.AbstractEventLoop
    ):
        pass

    def __dealloc__(self):
        if self.zpoller is not NULL:
            logger.warning('NodeActor.zpoller could not be deallocated')
        if self.zyre is not NULL:
            logger.warning('NodeActor.zyre could not be deallocated')
        if self.zactor_pipe is not NULL:
            logger.warning('NodeActor.zactor_pipe could not be deallocated')
        if self.zactor is not NULL:
            logger.warning('NodeActor.zactor could not be deallocated')

    def assert_lthread(self):
        assert threading.get_ident() == self.lthreadid, \
            '%s must be called from the loop thread' % sys._getframe(1).f_code.co_name

    def assert_zthread(self):
        assert threading.get_ident() == self.zthreadid, \
            '%s must be called from the zactor thread' % sys._getframe(1).f_code.co_name

    def start(self):
        """
        Start the node actor thread.

        This method is *not* thread safe and should only be called from the event loop thread.
        """
        self.assert_lthread()
        if self.started is not None:
            raise StopFailed('NodeActor already running')
        # Steal a reference to the future for the duration of zactor's run;
        # actor_run calls Py_DECREF on termination.
        Py_INCREF(self)

        self.started = futures.ThreadSafeFuture(loop=self.loop)
        self.stopped = futures.ThreadSafeFuture(loop=self.loop)

        with nogil:
            zactor = z.zactor_new(node_act, <void*>self)

        if zactor is NULL:
            Py_DECREF(self)
            raise MemoryError("Could not create zactor instance")

        self.zactor = zactor

    def stop(self):
        """
        Stop the node actor thread.

        This method is *not* thread safe and should only be called from the event loop thread.
        """
        self.assert_lthread()
        if self.started is None:
            raise StopFailed('NodeActor not running')
        with nogil:
            z.zactor_destroy(&self.zactor)
            self.zactor = NULL

        # We stole a reference to the future in NodeActor.start(), give it back
        Py_DECREF(self)

    def stop_sync(self):
        yield from self.stop().__await__()

    def give(self, fut: futures.ThreadSafeFuture):
        """
        Give a future for processing by the zactor thread.
        The future's result will be the corresponding zyre_* function's return value.

        This method is thread safe.
        """
        self.inbox.put(fut)
        self.loop.call_soon_threadsafe(self.signal_incoming)

    def take(self, timeout: int = None) -> messages.Msg:
        """
        Receive a future for processing by the zactor thread.

        This method is thread safe.
        """
        return self.inbox.get(timeout=timeout)

    def emit(self, msg: messages.Msg):
        """
        Emit an incoming zyre message.

        This method is thread safe.
        """
        self.loop.call_soon_threadsafe(self.outbox.put_nowait, msg)

    def signal_incoming(self):
        """
        Notify the zactor thread to check its inbox.

        This method is *not* thread safe and should only be called from the event loop thread.
        """
        self.assert_lthread()
        with nogil:
            # notify zactor's poller to check inbox
            z.zstr_send(self.zactor, signals.INCOMING)

    def configure(object self):
        """
        Configure and start the zyre node for this zactor.

        This method is *not* thread safe and should only be called from the zactor thread.
        """

        self.assert_zthread()
        name = self.config.name.encode('utf8')
        self.zyre = z.zyre_new(name)
        if self.zyre is NULL:
            raise MemoryError('Could not create zyre instance')

        if self.config.verbose:
            z.zyre_set_verbose(self.zyre)

        for k, v in self.config.headers.items():
            k = k.encode('utf8')
            v = v.encode('utf8')
            key = <char*>k
            value = <char*>v
            z.zyre_set_header(self.zyre, key, "%s", value)

        if self.config.interface:
            interface = self.config.interface.encode('utf8')
            interface = <char*>interface
            z.zyre_set_interface(self.zyre, interface)

        if self.config.endpoint:
            endpoint = self.config.endpoint.encode('utf8')
            endpoint = <char*>endpoint
            z.zyre_set_endpoint(self.zyre, "%s", endpoint)

        if self.config.gossip_endpoint:
            gossip_endpoint = self.config.gossip_endpoint.encode('utf8')
            gossip_endpoint = <char*>gossip_endpoint
            z.zyre_gossip_connect(self.zyre, "%s", gossip_endpoint)
            z.zyre_gossip_bind(self.zyre, "%s", gossip_endpoint)
            # Give some time for the bind/connect to occur
            z.zclock_sleep(250)

        if self.config.evasive_timeout_ms is not None:
            z.zyre_set_evasive_timeout(self.zyre, self.config.evasive_timeout_ms)

        if self.config.expired_timeout_ms is not None:
            z.zyre_set_expired_timeout(self.zyre, self.config.expired_timeout_ms)

        if z.zyre_start(self.zyre) != 0:
            z.zyre_destroy(&self.zyre)
            raise StartFailed('Could not start zyre instance')

        # Give some time for the start to occur
        z.zclock_sleep(250)

        self.zpoller = z.zpoller_new(self.zactor_pipe, NULL)
        if self.zpoller is NULL:
            z.zyre_destroy(&self.zyre)
            raise MemoryError('Could not create zpoller instance')

        z.zpoller_add(self.zpoller, z.zyre_socket(self.zyre))

        for g in self.config.groups:
            group = g.encode('utf8')
            group = <char*>group
            z.zyre_join(self.zyre, group)

        return 0

    def listen(self):
        """
        Listen for socket and inbox events and process them. Runs on the zactor thread until stopped.

        This method is *not* thread safe and should only be called from the zactor thread.
        """
        self.assert_zthread()
        cdef:
            int terminated = 0
            void * which
            char * cmd
            z.zmsg_t * zmsg
        with nogil:
            while not (terminated or z.zsys_interrupted):
                which = z.zpoller_wait(self.zpoller, -1)
                if which is z.zyre_socket(self.zyre):
                    zmsg = z.zmsg_recv(which)
                    if zmsg is NULL:
                        terminated = 1
                    else:
                        with gil:
                            msg = util.zmsg_to_msg(zmsg)
                            self.emit(msg)
                elif which is self.zactor_pipe:
                    cmd = z.zstr_recv(which)
                    if strcmp(cmd, signals.TERMINATE) == 0:
                        terminated = 1
                    elif strcmp(cmd, signals.INCOMING) == 0:
                        with gil:
                            self.process_inbox()
                    else:
                        with gil:
                            logger.error('node_actor_loop: received unknown cmd %s' % (<bytes>cmd).decode('utf8'))
                    free(cmd)
                if z.zpoller_terminated(self.zpoller):
                    terminated = 1

    def process_inbox(self):
        """
        Dequeue an item (future) from the inbox, process it, and set its result.

        This method is *not* thread safe and should only be called from the zactor thread.
        """
        self.assert_zthread()
        cdef:
            char * group
            char * peer
            char * blob
            char * address
            char * header
            char * value
            z.zlist_t * zlist
            int sig

        try:
            fut = self.take(timeout=5)
        except TimeoutError:
            logger.warning('Received signal but no message in inbox')
            return

        Py_INCREF(fut)
        try:
            sig = fut.signal
            if sig == signals.SHOUT:
                group = fut.group
                blob = fut.blob
                with nogil:
                    z.zyre_shouts(self.zyre, group, "%s", blob)
                fut.set_result(None)
            elif sig == signals.WHISPER:
                peer = fut.peer
                blob = fut.blob
                with nogil:
                    z.zyre_whispers(self.zyre, peer, "%s", blob)
                fut.set_result(None)
            elif sig == signals.JOIN:
                group = fut.group
                with nogil:
                    z.zyre_join(self.zyre, group)
                fut.set_result(None)
            elif sig == signals.LEAVE:
                group = fut.group
                with nogil:
                    z.zyre_leave(self.zyre, group)
                fut.set_result(None)
            elif sig == signals.PEERS:
                with nogil:
                    zlist = z.zyre_peers(self.zyre)
                if zlist is not NULL:
                    retset = util.zlist_to_str_set(zlist)
                    fut.set_result(retset)
                else:
                    fut.set_result(set())
            elif sig == signals.PEERS_BY_GROUP:
                group = fut.group
                with nogil:
                    zlist = z.zyre_peers_by_group(self.zyre, group)
                if zlist is not NULL:
                    retset = util.zlist_to_str_set(zlist)
                    fut.set_result(retset)
                else:
                    fut.set_result(set())
            elif sig == signals.OWN_GROUPS:
                with nogil:
                    zlist = z.zyre_own_groups(self.zyre)
                if zlist is not NULL:
                    retset = util.zlist_to_str_set(zlist)
                    fut.set_result(retset)
                else:
                    fut.set_result(set())
            elif sig == signals.PEER_GROUPS:
                with nogil:
                    zlist = z.zyre_peer_groups(self.zyre)
                if zlist is not NULL:
                    retset = util.zlist_to_str_set(zlist)
                    fut.set_result(retset)
                else:
                    fut.set_result(set())
            elif sig == signals.PEER_HEADER_VALUE:
                peer = fut.peer
                header = fut.header
                with nogil:
                    value = z.zyre_peer_header_value(self.zyre, peer, header)
                if value is not NULL:
                    fut.set_result((<bytes>value).decode('utf8'))
                    free(value)
                else:
                    fut.set_result(None)
            else:
                fut.set_exception(ValueError('Unknown signal'))
        except Exception as exc:
            fut.set_exception(exc)
            raise
        finally:
            Py_DECREF(fut)

    def act(self):
        """
        Long running function that handles inputs and outputs from zyre <-> Node.

        This is the entrypoint for the zactor thread.
        """
        try:
            self.assert_zthread()
            # Start and configure the zyre node
            self.configure()
            # Attach the zyre node's UUID to the actor
            self.uuid = (<bytes>z.zyre_uuid(self.zyre)).decode('utf8')
            self.uuid = (<bytes>z.zyre_uuid(self.zyre)).decode('utf8')
            # Notify zmq that the zactor is ready to start receiving
            z.zsock_signal(self.zactor_pipe, 0)
            # Notify NodeActor.start() that the zactor is ready to start receiving
            self.started.set_result(True)
        except Exception as exc:
            self.started.set_exception(exc)
            # We stole a reference to the future in NodeActor.start(), give it back
            Py_DECREF(self)
            return

        exc = None
        try:
            self.listen()
            # Notify any receivers we've stopped
            self.emit(Stopped())
        except Exception as e:
            logger.exception(e)
            exc = e
        finally:
            with nogil:
                if self.zpoller is not NULL:
                    z.zpoller_destroy(&self.zpoller)
                    self.zpoller = NULL
                    self.zactor_pipe = NULL
                if self.zyre is not NULL:
                    z.zyre_stop(self.zyre)
                    z.zclock_sleep(500)
                    z.zyre_destroy(&self.zyre)
                    self.zyre = NULL

            if exc:
                self.stopped.set_exception(exc)
            else:
                # Notify NodeActor.stop() we've stopped
                self.stopped.set_result(True)


cdef void node_act(z.zsock_t * pipe, void * _self) nogil:
    """
    Long running function that handles inputs and outputs from zyre <-> Node.

    This is the entrypoint for the zactor thread.
    """
    with gil:
        self = <NodeActor>_self
        self.zthreadid = threading.get_ident()
        self.zactor_pipe = pipe
        self.act()





import asyncio
import signal

from typing import Union, Mapping, Iterable, Set

from .exceptions import StartFailed, StopFailed

from . import futures
from . import nodeactor
from . import nodeconfig
from . import messages


class Node:
    __slots__ = ('config', 'loop', 'running', 'startstoplock', 'actor')

    def __init__(
        self,
        name: str, *,
        headers: Mapping = None,
        groups: Union[None, Iterable[str]] = None,
        endpoint: str = None,
        gossip_endpoint: str = None,
        interface: str = None,
        evasive_timeout_ms: int = 5000,
        expired_timeout_ms: int = 30000,
        verbose: bool = False,
        loop: asyncio.AbstractEventLoop = None
    ):
        """
        Constructor, creates a new Zyre node. Note that until you start the
        node it is silent and invisible to other nodes on the network.
        The node name is provided to other nodes during discovery.
        """
        self.actor = None
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.startstoplock = asyncio.Lock()
        self.config = nodeconfig.NodeConfig(
            name=name, headers=headers, groups=groups, endpoint=endpoint, gossip_endpoint=gossip_endpoint,
            interface=interface, evasive_timeout_ms=evasive_timeout_ms, expired_timeout_ms=expired_timeout_ms,
            verbose=verbose
        )
        self.running = False

    @property
    def name(self):
        return self.config.name

    @property
    def uuid(self):
        return self.actor.uuid

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash(self.uuid)

    async def start(self):
        """
        Start node, after setting header values. When you start a node it
        begins discovery and connection. Returns 0 if OK, -1 if it wasn't
        possible to start the node.
        """
        async with self.startstoplock:
            if self.running:
                raise StartFailed('Node already running')
            self.actor = nodeactor.NodeActor(config=self.config, loop=self.loop)
            self.loop.add_signal_handler(signal.SIGINT, self.stop_sync)
            self.loop.add_signal_handler(signal.SIGABRT, self.stop_sync)
            self.actor.start()
            await self.actor.started
            self.running = True

    async def stop(self):
        """
        Stop node; this signals to other peers that this node will go away.
        This is polite; however you can also just destroy the node without
        stopping it.
        """
        async with self.startstoplock:
            if not self.running:
                raise StopFailed('Node not running')
            self.actor.stop()
            await self.actor.stopped
            self.running = False
            self.loop.remove_signal_handler(signal.SIGINT)
            self.loop.remove_signal_handler(signal.SIGABRT)

    def stop_sync(self):
        yield from self.stop().__await__()

    async def recv(self, timeout: int = None) -> messages.Msg:
        """
        Receive next message from network; the message may be a control
        message (ENTER, EXIT, JOIN, LEAVE) or data (WHISPER, SHOUT).
        Returns Msg object, or NULL if interrupted

        Note that having multiple tasks consuming from recv() will result in
        skipped messages, as each recv() task will destructively pop items from the queue.

        This method is *not* thread safe and should only be called from the event loop thread.
        """
        msg = await asyncio.wait_for(asyncio.ensure_future(self.actor.outbox.get()), timeout=timeout)
        self.actor.outbox.task_done()
        if isinstance(msg, Exception):
            raise msg
        return msg

    async def shout(self, group: str, blob: Union[bytes, str]):
        """
        Send message to a group
        """
        if isinstance(blob, str):
            blob = blob.encode('utf8')
        fut = futures.ShoutFuture(group=group, blob=blob, loop=self.loop)
        self.actor.give(fut)
        await asyncio.ensure_future(fut)

    async def whisper(self, peer: str, blob: Union[bytes, str]):
        """
        Send message to single peer, specified as a UUID string
        """
        if isinstance(blob, str):
            blob = blob.encode('utf8')
        fut = futures.WhisperFuture(peer=peer, blob=blob, loop=self.loop)
        self.actor.give(fut)
        await asyncio.ensure_future(fut)

    async def join(self, group: str):
        """
        Join a named group; after joining a group you can send messages to
        the group and all Zyre nodes in that group will receive them.
        """
        fut = futures.JoinFuture(group=group, loop=self.loop)
        self.actor.give(fut)
        await asyncio.ensure_future(fut)

    async def leave(self, group: str):
        """
        Leave a named group
        """
        fut = futures.LeaveFuture(group=group, loop=self.loop)
        self.actor.give(fut)
        await asyncio.ensure_future(fut)

    async def peers(self) -> Set[str]:
        """
        Return set of current peer ids.
        """
        fut = futures.PeersFuture(loop=self.loop)
        self.actor.give(fut)
        return await asyncio.ensure_future(fut)

    async def peers_by_group(self, group: str) -> Set[str]:
        """
        Return set of current peers of this group.
        """
        fut = futures.PeersByGroupFuture(group=group, loop=self.loop)
        self.actor.give(fut)
        return await asyncio.ensure_future(fut)

    async def own_groups(self) -> Set[str]:
        """
        Return set of currently joined groups.
        """
        fut = futures.OwnGroupsFuture(loop=self.loop)
        self.actor.give(fut)
        return await asyncio.ensure_future(fut)

    async def peer_groups(self) -> Set[str]:
        """
        Return set of groups known through connected peers.
        """
        fut = futures.PeerGroupsFuture(loop=self.loop)
        self.actor.give(fut)
        return await asyncio.ensure_future(fut)

    async def peer_header_value(self, peer: str, header: str) -> str:
        """
        Return the value of a header of a connected peer.
        Returns null if peer or key doesn't exits.
        """
        fut = futures.PeerHeaderValueFuture(peer=peer, header=header, loop=self.loop)
        self.actor.give(fut)
        return await asyncio.ensure_future(fut)




import asyncio


_SHOUT = 0
_WHISPER = 1
_JOIN = 2
_LEAVE = 3
_PEERS = 4
_PEERS_BY_GROUP = 5
_OWN_GROUPS = 6
_PEER_GROUPS = 7
_PEER_ADDRESS = 8
_PEER_HEADER_VALUE = 9


class ThreadSafeFuture:
    _asyncio_future_blocking = True

    def __init__(self, *, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self.future = self._loop.create_future()

    def result(self):
        """
        Return the result this future represents.

        If the future has been cancelled, raises CancelledError.  If the
        future's result isn't yet available, raises InvalidStateError.  If
        the future is done and has an exception set, this exception is raised.
        """
        return self.future.result()

    def set_result(self, result):
        """
        Set the future result.

        This method can be called from any thread but is not guaranteed to set the result immediately.
        """
        self._loop.call_soon_threadsafe(self.future.set_result, result)

    def cancel(self, *args, **kwargs):
        """
        Cancel the future and schedule callbacks.

        If the future is already done or cancelled, return False.  Otherwise,
        change the future's state to cancelled, schedule the callbacks and
        return True.

        This method can be called from any thread but is not guaranteed to cancel the future immediately.
        """
        self._loop.call_soon_threadsafe(self.future.cancel)

    def cancelled(self):
        """ Return True if the future was cancelled. """
        return self.future.cancelled()

    def done(self):
        """
        Return True if the future is done.

        Done means either that a result / exception are available, or that the
        future was cancelled.
        """
        return self.future.done()

    def exception(self):
        """
        Return the exception that was set on this future.

        The exception (or None if no exception was set) is returned only if
        the future is done.  If the future has been cancelled, raises
        CancelledError.  If the future isn't done yet, raises
        InvalidStateError.
        """
        return self.future.exception()

    def get_loop(self):
        """ Return the event loop the Future is bound to. """
        return self.future.get_loop()

    def add_done_callback(self, callback):
        """
        Add a callback to be run when the future becomes done.

        The callback is called with a single argument - the future object. If
        the future is already done when this is called, the callback is
        scheduled with call_soon.

        This method can be called from any thread but is not guaranteed to add the callback immediately.
        """
        self._loop.call_soon_threadsafe(self.future.add_done_callback, callback)

    def remove_done_callback(self, callback):
        """
        Remove all instances of a callback from the "call when done" list.

        Returns the number of callbacks removed.

        This method can be called from any thread but is not guaranteed to remove the callback immediately.
        """
        self._loop.call_soon_threadsafe(self.future.remove_done_callback, callback)

    def set_exception(self, exception):
        """
        Mark the future done and set an exception.

        If the future is already done when this method is called, raises
        InvalidStateError.

        This method can be called from any thread but is not guaranteed to set the exception immediately.
        """
        self._loop.call_soon_threadsafe(self.future.set_exception, exception)

    def __await__(self):
        return self.future.__await__()

    def __del__(self):
        return self.future.__del__()

    def __iter__(self):
        return self.future.__iter__()


class StartedFuture(ThreadSafeFuture):
    pass


class SignalFuture(ThreadSafeFuture):
    pass


class ShoutFuture(SignalFuture):
    signal = _SHOUT

    def __init__(self, *, group: str, blob: bytes, **kwargs):
        self.group = group.encode('utf8')
        self.blob = blob
        super().__init__(**kwargs)


class WhisperFuture(SignalFuture):
    signal = _WHISPER

    def __init__(self, *, peer: str, blob: bytes, **kwargs):
        self.peer = peer.encode('utf8')
        self.blob = blob
        super().__init__(**kwargs)


class JoinFuture(SignalFuture):
    signal = _JOIN

    def __init__(self, *, group: str, **kwargs):
        self.group = group.encode('utf8')
        super().__init__(**kwargs)


class LeaveFuture(SignalFuture):
    signal = _LEAVE

    def __init__(self, *, group: str, **kwargs):
        self.group = group.encode('utf8')
        super().__init__(**kwargs)


class PeersFuture(SignalFuture):
    signal = _PEERS

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class PeersByGroupFuture(SignalFuture):
    signal = _PEERS_BY_GROUP

    def __init__(self, *, group: str, **kwargs):
        self.group = group.encode('utf8')
        super().__init__(**kwargs)


class OwnGroupsFuture(SignalFuture):
    signal = _OWN_GROUPS

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class PeerGroupsFuture(SignalFuture):
    signal = _PEER_GROUPS

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class PeerHeaderValueFuture(SignalFuture):
    signal = _PEER_HEADER_VALUE

    def __init__(self, *, peer: str, header: str, **kwargs):
        self.peer = peer.encode('utf8')
        self.header = header.encode('utf8')
        super().__init__(**kwargs)




# cython: language_level=3

"""
Cython bindings for Zyre and various related zmq/czmq utilities
"""


from libc.stdint cimport uint64_t
from libcpp cimport bool

ctypedef unsigned char byte


cdef extern from "zyre.h" nogil:

    # zsys.h

    int zsys_interrupted

    # zclock.h

    void zclock_sleep (int msecs)

    # zsock.h

    ctypedef struct zsock_t

    int zsock_signal (void * self, byte status)

    void zsock_destroy (zsock_t **self_p)

    # zlist.h

    ctypedef struct zlist_t

    void zlist_destroy(zlist_t ** self_p)

    void * zlist_pop(zlist_t * self)

    size_t zlist_size(zlist_t * self)

    bool zlist_exists (zlist_t *self, void *item)

    # zmsg.h

    ctypedef struct zmsg_t

    void zmsg_destroy(zmsg_t ** self_p)

    zmsg_t * zmsg_recv (void *source)

    size_t zmsg_size(zmsg_t * self)

    char * zmsg_popstr(zmsg_t * self)

    # zstr.h

    int zstr_send (void *dest, const char *string)

    char * zstr_recv (void *source)

    # zactor.h
    ctypedef void zactor_fn(zsock_t * pipe, void * args)

    ctypedef struct zactor_t

    zactor_t * zactor_new(zactor_fn actor, void * args)

    void zactor_destroy(zactor_t ** self_p)

    # zpoller.h

    ctypedef struct zpoller_t

    zpoller_t * zpoller_new(void * reader, ...)

    void zpoller_destroy(zpoller_t** self_p)

    int zpoller_add(zpoller_t * self, void * reader)

    int zpoller_remove (zpoller_t *self, void *reader)

    void * zpoller_wait(zpoller_t * self, int timeout)

    bool zpoller_terminated(zpoller_t * self)

    # zyre.h

    ctypedef struct zyre_t

    zyre_t * zyre_new(char * name)

    void zyre_destroy(zyre_t** self_p)

    char * zyre_uuid(zyre_t * self)

    char * zyre_name(zyre_t * self)

    void zyre_set_header(zyre_t * self, char * name, char * fmt, char * value)

    void zyre_set_verbose(zyre_t * self)

    void zyre_set_evasive_timeout(zyre_t * self, int interval)

    void zyre_set_expired_timeout(zyre_t * self, int interval)

    void zyre_set_interface(zyre_t * self, char * value)

    int zyre_set_endpoint(zyre_t * self, char * fmt, char * value)

    void zyre_set_port(zyre_t *self, int port_nbr)

    void zyre_gossip_bind(zyre_t * self, char * fmt, char * value)

    void zyre_gossip_connect(zyre_t * self, char * fmt, char * value)

    int zyre_start(zyre_t * self)

    void zyre_stop(zyre_t * self)

    int zyre_join(zyre_t * self, char * group)

    int zyre_leave(zyre_t * self, char * group)

    zmsg_t * zyre_recv(zyre_t * self)

    int zyre_whisper(zyre_t * self, char * peer, zmsg_t** msg_p)

    int zyre_shout(zyre_t * self, char * group, zmsg_t** msg_p)

    int zyre_whispers(zyre_t * self, char * peer, char * fmt, char * value)

    int zyre_shouts(zyre_t * self, char * group, char * fmt, char * value)

    zlist_t * zyre_peers(zyre_t * self)

    zlist_t * zyre_peers_by_group(zyre_t * self, char * name)

    zlist_t * zyre_own_groups(zyre_t * self)

    zlist_t * zyre_peer_groups(zyre_t * self)

    char * zyre_peer_address(zyre_t * self, char * peer)

    char * zyre_peer_header_value(zyre_t * self, char * peer, char * name)

    zsock_t * zyre_socket(zyre_t * self)

    uint64_t zyre_version()



# cython: language_level=3

cdef enum SIGNALS:
    SHOUT, WHISPER, JOIN, LEAVE, PEERS, PEERS_BY_GROUP, OWN_GROUPS, PEER_GROUPS, PEER_ADDRESS, PEER_HEADER_VALUE

cdef const char * TERMINATE
cdef const char * INCOMING



class Msg:
    __slots__ = ('event', 'peer', 'name', 'headers', 'address', 'group', 'blob')

    def __init__(
        self,
        *,
        event: str = None,
        peer: str = None,
        name: str = None,
        headers: str = None,
        address: str = None,
        group: str = None,
        blob: bytes = None
    ):
        self.event = event or ''
        self.peer = peer or ''
        self.name = name or ''
        self.headers = headers or ''
        self.address = address or ''
        self.group = group or ''
        self.blob = blob or b''

    def __repr__(self):
        args = ['{}={}'.format(slot, repr(getattr(self, slot))) for slot in self.__slots__]
        return '{}({})'.format(self.__class__.__name__, ", ".join(args))

    @property
    def string(self):
        return self.blob.decode('utf8') if self.blob is not None else None

    def to_dict(self):
        return dict(
            event=self.event,
            peer=self.peer,
            name=self.name,
            headers=self.headers,
            address=self.address,
            group=self.group,
            blob=self.blob,
        )



from typing import Mapping, Union, Iterable


class NodeConfig:
    def __init__(
        self,
        name: str, * ,
        headers: Mapping = None,
        groups: Union[None, Iterable[str]] = None,
        endpoint: str = None,
        gossip_endpoint: str = None,
        interface: str = None,
        evasive_timeout_ms: int = 5000,
        expired_timeout_ms: int = 30000,
        verbose: bool = False
    ):
        self.name = name
        self.headers = headers or {}
        self.groups = groups or set()
        self.endpoint = endpoint
        self.gossip_endpoint = gossip_endpoint
        self.interface = interface
        self.evasive_timeout_ms = evasive_timeout_ms
        self.expired_timeout_ms = expired_timeout_ms
        self.verbose = int(verbose)




from setuptools import setup, Extension


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='aiozyre',
    version='1.1.3',
    description='asyncio-friendly Python bindings for Zyre',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Elijah Shaw-Rutschman',
    author_email='elijahr+aiozyre@gmail.com',
    packages=['aiozyre'],
    package_dir={
      'aiozyre': 'src/aiozyre',
    },
    package_data={'aiozyre': [
        '*.pyx',
        '*.pxd',
    ]},
    ext_modules=[
        Extension('aiozyre.nodeactor', sources=['src/aiozyre/nodeactor.pyx'], libraries=['czmq', 'zyre']),
        Extension('aiozyre.signals', sources=['src/aiozyre/signals.pyx']),
        Extension('aiozyre.util', sources=['src/aiozyre/util.pyx'], libraries=['czmq', 'zyre']),
    ],
    setup_requires=['cython'],
    extras_require={
        'dev': [
            'blessed',
            'aioconsole',
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3',
        'Topic :: System :: Networking',
        'Framework :: AsyncIO',
    ],
)



# cython: language_level=3

from . cimport zyre as z


cdef class NodeActor:
    cpdef public str uuid
    cpdef public object started
    cpdef public object stopped
    cpdef public object config
    cpdef public object loop
    cpdef public object inbox
    cpdef public object outbox

    # private
    cdef z.zyre_t * zyre
    cdef z.zpoller_t * zpoller
    cdef z.zactor_t * zactor
    cdef z.zsock_t * zactor_pipe
    cpdef unsigned long zthreadid
    cpdef unsigned long lthreadid


cdef void node_act(z.zsock_t * pipe, void * _actor) nogil
